/* fsgrab.c -- get blocks from an (ext2) filesystem on Linux
 *
 * Aaron Crane <aaronc@pobox.com>
 *
 * This file is part of `e2recover', a suite of programs which assist in
 * recovering deleted files from ext2 file systems.  `e2recover' is
 * Copyright  1997, 1998, 1999 Aaron Crane <aaronc@pobox.com>.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 *
 * 1997-10-17: v1.1
 *
 *   Improved program structure.
 *   Now reads until error or eof or found enough data.
 *   Now writes until error or written enough data.
 *   Added `-f' (force stdout writes) option.
 *
 * 1998-09-10: v1.2
 *
 *   Autoconf-ized.
 *   Also works on systems without gcc, gmake, glibc
 *   Attempt to use ext2fs_llseek or llseek where available.
 *
 * 1999-01-15: e2recover 1.0 (formerly fsgrab 1.3)
 *
 *   Major reorganisation.
 *   Improved error checking.
 *   Use a robust parsing routine for numbers.
 *   Added -m, -o, -a options.
 *   Broke some code out into a separate file.
 *   Long options and other changes to conform to GNU standards.texi.
 */

#include <misc.h>

#include <getopt.h>
#include <unistd.h>
#include <sys/types.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>
#include <stdlib.h>
#include <limits.h>

#ifndef SSIZE_MAX
#define SSIZE_MAX (INT_MAX)
#endif

static struct option const long_opts[] =
{
    { "device",     required_argument, 0, 'd' },
    { "block-size", required_argument, 0, 'b' },
    { "blocksize",  required_argument, 0, 'b' },
    { "count",      required_argument, 0, 'c' },
    { "skip",       required_argument, 0, 's' },
    { "max-chars",  required_argument, 0, 'm' },
    { "max-bytes",  required_argument, 0, 'm' },
    { "output",     required_argument, 0, 'o' },
    { "append",     no_argument,       0, 'a' },
    { "force",      no_argument,       0, 'f' },
    { "help",       no_argument,       0, 'h' },
    { "version",    no_argument,       0, 'V' },
    { NULL,         0,                 0,  0  }
};

struct output_info
{
    const char *name;
    int fd;
    unsigned append:1;
    unsigned allow_ttys:1;
};

struct request_info
{
    const char *name;
    int fd;
    size_t blocksize, count, skip, maxlen;
    unsigned truncate:1;
};

struct buffer
{
    unsigned char *data;
    size_t n;
};

static void rq_init (struct request_info *rq);
static void rq_check (struct request_info *rq);
static void rq_open (struct request_info *rq);
static void rq_seek (const struct request_info *rq);
static void rq_read (const struct request_info *rq, struct output_info *out, struct buffer *buf);
static void rq_close (struct request_info *rq);

static void out_init (struct output_info *out);
static void out_check (struct output_info *out);
static void out_open (struct output_info *out);
static void out_write (const struct request_info *rq, struct output_info *out, struct buffer *buf);
static void out_close (struct output_info *out);

static int fsgrab (struct output_info *out, struct request_info *rq);

static void /*@exits@*/ long_usage (void) __attribute__ ((noreturn));
static void /*@exits@*/ show_version (void) __attribute__ ((noreturn));

static void
long_usage (void)
{
    printf ("Usage: %s [OPTION]... [DEVICE]\n"
	    "\n"
	    "Grab blocks from DEVICE (usually a block device), or /dev/hda1 by default.\n"
	    "Output is sent to standard output unless an OUTPUT-FILE is given.\n"
	    "\n"
	    "  -d, --device=DEVICE          read from DEVICE (default /dev/hda1)\n"
	    "  -b, --block-size=BLOCKSIZE   read blocks of BLOCKSIZE bytes (default 1024)\n"
	    "  -c, --count=COUNT            read COUNT blocks (default 1)\n"
	    "  -s, --skip=SKIP              seek past SKIP blocks before reading\n"
	    "  -m, --max-chars=MAXIMUM,\n"
	    "      --max-bytes=MAXIMUM      write no more than MAXIMUM bytes\n"
	    "  -o, --output=OUTPUT-FILE     write to OUTPUT instead of stdout\n"
	    "  -a, --append                 when using -o, open OUTPUT in append mode\n"
	    "  -f, --force                  allow writing to stdout even if it's a terminal\n"
	    "      --help                   display this help and exit\n"
	    "      --version                display version information and exit\n"
	    "\n"
	    "BLOCKSIZE and MAXIMUM may have an optional multiplier suffix: w for 2,\n"
	    "b for 512, k for 1024, m for 1Meg.\n"
	    "\n"
	    "Specifying a -d DEVICE or --device=DEVICE option prohibits specifying a\n"
	    "device as a non-option argument\n"
	    "\n"
	    "Report bugs to <aaronc@pobox.com>.\n", invocation_name);
    exit (0);
}

void
short_usage (void)
{
    fprintf (stderr, "Try `%s --help' for more information.\n", invocation_name);
    exit (EXIT_FAILURE);
}

static void
show_version (void)
{
    printf ("fsgrab (%s %s)\n"
	    "Copyright (C) 1998 Aaron Crane\n"
	    "Copyright (C) 1997 Free Software Foundation, Inc.\n"
	    "fsgrab comes with NO WARRANTY, to the extent permitted by law.\n"
	    "You may modify and redistribute copies of fsgrab under the terms of the\n"
	    "GNU General Public License.  For more information about these matters,\n"
	    "see the file named COPYING.\n",
	    PACKAGE, VERSION);
    exit (0);
}


int
main (int argc, char **argv)
{
    struct output_info out;
    struct request_info rq;
    int opt;
    int allow_nonopt = 1;

    invocation_name = argv[0];
    if (!invocation_name || !invocation_name[0])
	invocation_name = "fsgrab";

    rq_init (&rq);
    out_init (&out);

    while ((opt = getopt_long (argc, argv, "b:c:d:m:o:s:afhV", long_opts, 0)) != EOF)
	switch (opt)
	{
	case 'b':
	    rq.blocksize = safe_atosz (optarg, 1);
	    break;
	case 'c':
	    rq.count = safe_atosz (optarg, 0);
	    break;
	case 'd':
	    allow_nonopt = 0;
	    rq.name = optarg;
	    break;
	case 'm':
	    rq.truncate = 1;
	    rq.maxlen = safe_atosz (optarg, 1);
	    break;
	case 'o':
	    out.name = optarg;
	    break;
	case 's':
	    rq.skip = safe_atosz (optarg, 0);
	    break;
	case 'a':
	    out.append = 1;
	    break;
	case 'f':
	    out.allow_ttys = 1;
	    break;
	case 'h':
	    long_usage ();
	case 'V':
	    show_version ();
	default:
	    short_usage ();
	}

    if (optind < argc - 1)
    {
	message ("too many non-option arguments\n");
	short_usage ();
    }
    else if (optind == argc - 1)
    {
	if (allow_nonopt)
	    rq.name = argv[optind];
	else
	{
	    message ("invalid non-option argument\n");
	    short_usage ();
	}
    }

    return fsgrab (&out, &rq) == -1;
}


static void
rq_init (struct request_info *rq)
{
    rq->name = "/dev/hda1";
    rq->fd = -1;
    rq->blocksize = 1024;
    rq->count = 1;
    rq->skip = 0;
    rq->maxlen = 0;
    rq->truncate = 0;
}

static void
rq_check (struct request_info *rq)
{
    size_t rq_size = rq->count * rq->blocksize;
    if (!rq->truncate)
	rq->maxlen = rq_size;
    else
    {
	if (rq->maxlen > rq_size)
	    message ("warning: maximum length (%lu) is greater than input size (%lu)\n",
		     (unsigned long) rq->maxlen, (unsigned long) rq_size);
	if (rq_size - rq->maxlen > rq->blocksize)
	    message ("warning: maximum length (%lu) truncates by %lu, "
		     "which is more than block size (%lu)\n",
		     (unsigned long) rq->maxlen,
		     (unsigned long) (rq_size - rq->maxlen),
		     (unsigned long) rq->blocksize);
    }
}

static void
rq_open (struct request_info *rq)
{
    rq->fd = open (rq->name, O_RDONLY);
    if (rq->fd == -1)
	fatal ("couldn't open %s: %s\n", rq->name, strerror (errno));
}

static void
rq_seek (const struct request_info *rq)
{
    my_offset_t seekpos = (my_offset_t) rq->blocksize * rq->skip;

    if (seekpos > 0)
    {
	my_offset_t i = SEEK (rq->fd, seekpos, SEEK_SET);
	if (i == -1 || i != seekpos)
	    fatal ("%s: " SEEKNAME " failed: %s\n", rq->name, strerror (errno));
    }
}

static void
rq_read (const struct request_info *rq, struct output_info *out, struct buffer *buf)
{
    size_t n = rq->count * rq->blocksize;
    unsigned char *p = buf->data;

    /* safe to cast SSIZE_MAX to size_t */
    if (n > (size_t) SSIZE_MAX
	|| n / rq->blocksize != rq->count)
	fataln (2, "FIXME: won't read more bytes than an ssize_t can enumerate\n");

    while (n > 0)
    {
	ssize_t i = read (rq->fd, p, n);
	if (i < -1)
	    fataln (2, "%s: read %ld bytes -- can't happen; errno is %s\n",
		    rq->name, (long) i, strerror (errno));
	else if (i == -1)
	{
	    message ("%s: read failed: %s; trying to write the %lu chars I've got\n",
		     rq->name, strerror (errno), (unsigned long) (p - buf->data));
	    /* last-ditch attempt to write what we've got */
	    out_write (rq, out, buf);
	    exit (1);
	}
	else if (i == 0)
	{
	    message ("%s: warning: unexpected end of file after %lu chars\n",
		     rq->name, (unsigned long) (p - buf->data));
	    break;
	}
	else			/* (i > 0) -- success */
	{
	    n -= (size_t) i;	/* safe to cast positive ssize_t to size_t */
	    p += i;
	}
    }

    /* Safe to assign ptrdiff_t (no, LCLint, it's not an int) to size_t,
     * because we can't have read more than N chars into buffer, and N is a
     * size_t. */
    buf->n = p - buf->data;
}

static void
rq_close (struct request_info *rq)
{
    if (close (rq->fd) == -1)
	message ("warning: can't close %s: %s\n", rq->name, strerror (errno));
    rq->fd = -1;
}


static void
out_init (struct output_info *out)
{
    out->name = 0;
    out->fd = -1;
    out->append = 0;
    out->allow_ttys = 0;
}

static void
out_check (struct output_info *out)
{
    if (out->name)
    {
	if (out->allow_ttys)	/* -o and -f are mutually exclusive */
	    long_usage ();
    }
    else
    {
	out->fd = 1;
	out->name = "<stdout>";
	if (out->append)
	    message ("warning: append mode is ignored when writing to stdout\n");
	if (!out->allow_ttys && isatty (out->fd))
	    fatal ("%s: stdout is a terminal\n", ttyname (out->fd));
    }
}

static void
out_open (struct output_info *out)
{
    static const mode_t open_mode[2] = { O_CREAT|O_TRUNC, O_APPEND };
    if (out->fd == -1)
    {
	out->fd = open (out->name, O_WRONLY | open_mode[out->append], 0666);
	if (out->fd == -1)
	    fatal ("couldn't open %s for writing: %s\n", out->name, strerror (errno));
    }
}

static void
out_write (const struct request_info *rq, struct output_info *out, struct buffer *buf)
{
    size_t n = buf->n;
    const unsigned char *p = buf->data;

    if (rq->truncate && n > rq->maxlen)
	n = rq->maxlen;

    while (n > 0)
    {
	ssize_t i = write (out->fd, p, n);
	if (i < -1)
	    fataln (2, "%s: wrote %ld bytes -- can't happen; errno is %s\n",
		    out->name, (long) i, strerror (errno));
	else if (i == -1)
	    fatal ("%s: write failed: %s\n", out->name, strerror (errno));
	else			/* (i >= 0) */
	{
	    /* We ignore the case where zero bytes were written -- we'll get
	     * an error on the next write. */
	    n -= (size_t) i;	/* safe to cast non-negative ssize_t to size_t */
	    p += i;
	}
    }
}

static void
out_close (struct output_info *out)
{
    if (close (out->fd) == -1)
	message ("warning: can't close %s: %s\n", out->name, strerror (errno));
    out->fd = -1;
}


static int
fsgrab (struct output_info *out, struct request_info *rq)
{
    struct buffer buf;

    rq_check (rq);
    out_check (out);

    buf.data = mustmalloc (rq->count * rq->blocksize);
    buf.n = 0;

    rq_open (rq);
    rq_seek (rq);
    rq_read (rq, out, &buf);
    rq_close (rq);

    out_open (out);
    out_write (rq, out, &buf);
    out_close (out);

    free (buf.data);

    return 0;
}
